package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor, TensorSegment}

import scala.collection.mutable

/**
  * The trait represents the split operation which can be abstracted away from the specific dimension through the TensorSegments
  */
trait AbstractSplit {

  /**
   * Splits the group with the highest entropy which is splittable.
    * As described in the paper for the split operation
   */
  def performOperation[X, Y, Z](dm: DataTensor[X, Y, Z]): DataTensor[X, Y, Z] = {
    performGroupSplit(dm, indicesOfMaxEncoding(dm))
  }

  protected def indicesOfMaxEncoding[X, Y, Z](dm: DataTensor[X, Y, Z]): Seq[Int]

  protected def performGroupSplit[X, Y, Z](dm: DataTensor[X, Y, Z], splitSeq: Seq[Int]): DataTensor[X, Y, Z]

  protected def determineNewGroupIds[A, X, Y, Z](groupToBeSplit: TensorSegment[A, X, Y, Z]): (Set[A], Set[A]) = {
    val keepSet = mutable.Set[A]()
    val newSet = mutable.Set[A]()

    //initial nr of row-elements
    val initialRowCount = groupToBeSplit.groupIds.size
    var currentEncodingCosts = 0d
    var currentTensorProps = for (a <- groupToBeSplit.dimAIndices) yield for (b <- groupToBeSplit.dimBIndices) yield {
      val ca = groupToBeSplit.caCell(a, b)
      val ctp = CurrentCellProps(determineEntryMatrixSize(ca), initialRowCount, ca.trueCount)
      currentEncodingCosts += ctp.currentAvgEncodingCostsPerRow
      ctp
    }

    for {
      entryIdx <- groupToBeSplit.groupIds
    } {
      var removedEncodingCosts = 0d
      val tensorPropsWithRemovedRowEntry = for {
        a <- groupToBeSplit.dimAIndices} yield for {
        b <- groupToBeSplit.dimBIndices} yield {
        val upCtp = currentTensorProps(a)(b).removeRowElement(groupToBeSplit.getTrueCountForEntry(a, b)(entryIdx))
        removedEncodingCosts += upCtp.currentAvgEncodingCostsPerRow
        upCtp
      }

      if (removedEncodingCosts < currentEncodingCosts) {
        newSet += entryIdx
        currentTensorProps = tensorPropsWithRemovedRowEntry
        currentEncodingCosts = removedEncodingCosts
      } else {
        keepSet += entryIdx
      }
    }
    (keepSet.toSet, newSet.toSet)
  }

  //Determines CurrentTensorProps(nrOfMatrixElements: Int, //Fixed as only the number of row elements changes
  protected def determineEntryMatrixSize[X,Y,Z](caTensor: CaCell[X,Y,Z]) : Int
}
