package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor}

import scala.math._
/**
  * Fuses the two most similar columns (based on the average distance between encoding costs of true-entries over all tensor groups)
  */
object FusionCols {
   def performOperation[X, Y, Z](dm: DataTensor[X, Y, Z]): DataTensor[X, Y, Z] = {
    if (dm.nrOf2DGroups > 1) {
      val (idxA, idxB) = minEncodingPair(dm)

      val aIds = dm.getDim2Ids(idxA)
      val bIds = dm.getDim2Ids(idxB)
      val keep = min(idxA, idxB)
      val ignore = max(idxA, idxB)

      val combinedIds = aIds ++ bIds

      val newRowData = new Array[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]](dm.nrOf1DGroups)
      for {
        rowIdx <- dm.dim1Indices
      } {
        var currentIdxShift = 0
        val newColData = new Array[IndexedSeq[CaCell[X, Y, Z]]](dm.nrOf2DGroups - 1)
        for {
          colIdx <- dm.dim2Indices
        } {
          if (colIdx == keep) {
            newColData(colIdx) = dm.caCells(rowIdx)(colIdx).map { gt =>
              gt.copyWithNewDim2(combinedIds)
            }
          } else if (colIdx == ignore) {
            currentIdxShift += 1
          } else {
            newColData(colIdx - currentIdxShift) = dm.caCells(rowIdx)(colIdx)
            //Ignore the  case and from no on we have to shift the indices of all other cols by one
          }
        }
        newRowData(rowIdx) = newColData
      }
       dm.copyWithNewCaCells(newRowData)
    } else {
      dm
    }
  }

  def minEncodingPair[X, Y, Z](dm: DataTensor[X, Y, Z]): (Int, Int) = {
    val best = (0 until dm.nrOf2DGroups).view.combinations(2).maxBy { case Seq(xIdx, yIdx) =>
      val x = dm.getCaCellsInDim2(xIdx)
      val y = dm.getCaCellsInDim2(yIdx)
      val xCosts = x.segmentEncodingCosts
      val yCosts = y.segmentEncodingCosts
      val combinedCosts = x.combinedSegmentEncodingCosts(y)
      (xCosts + yCosts) - combinedCosts
    }
    best(0) -> best(1)
  }
}