package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor}

import scala.math._

/**
  * Fuses the two most similar tubes (based on the average distance between encoding costs of true-entries over all tensor groups)
  */
object FusionTubes  {
   def performOperation[X, Y, Z](dm: DataTensor[X, Y, Z]): DataTensor[X, Y, Z] = {
    if (dm.nrOf3DGroups > 1) {
      val (idxA, idxB) = minEncodingPair(dm)

      val aIds = dm.getDim3Ids(idxA)
      val bIds = dm.getDim3Ids(idxB)
      val keep = min(idxA, idxB)
      val ignore = max(idxA, idxB)

      val combinedIds = aIds ++ bIds

      val newRowData = new Array[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]](dm.nrOf1DGroups)
      for {
        rowIdx <- dm.dim1Indices
      } {
        val newColData = new Array[IndexedSeq[CaCell[X, Y, Z]]](dm.nrOf2DGroups)
        for {
          colIdx <- dm.dim2Indices
        } {
          var currentIdxShift = 0
          val newTubeData = new Array[CaCell[X, Y, Z]](dm.nrOf3DGroups - 1)
          for {
            tubeIdx <- dm.dim3Indices
          } {
            val old = dm.getCaCells(rowIdx, colIdx, tubeIdx)
            if (keep == tubeIdx) {
              newTubeData(tubeIdx) = {
                old.copyWithNewDim3(dim3Ids = combinedIds)
              }
            } else if (tubeIdx == ignore) {
              //we ignore the keep == tubeIdx since we combine it to the keep case
              currentIdxShift += 1
            } else {
              newTubeData(tubeIdx - currentIdxShift) = old
            }
          }
          newColData(colIdx) = newTubeData
        }
        newRowData(rowIdx) = newColData
      }
      dm.copyWithNewCaCells(newRowData)
    } else {
      dm
    }
  }

  def minEncodingPair[X, Y, Z](dm: DataTensor[X, Y, Z]): (Int, Int) = {
    val best = (0 until dm.nrOf3DGroups).view.combinations(2).maxBy { case Seq(xIdx, yIdx) =>
      val x = dm.getCaCellsInDim3(xIdx)
      val y = dm.getCaCellsInDim3(yIdx)
      val xCosts = x.segmentEncodingCosts
      val yCosts = y.segmentEncodingCosts
      val combinedCosts = x.combinedSegmentEncodingCosts(y)
      (xCosts + yCosts) - combinedCosts
    }
    best(0) -> best(1)
  }
}