package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor}


/**
  * Implements the regroup operation for columns
  */
object ReGroupCols extends  AbstractRegroup {
  protected def determineEntryMatrixSize[X, Y, Z](ca: CaCell[X, Y, Z]): Int = ca.matrixSizeForDim2Idx

   def performOperation[X, Y, Z](dm: DataTensor[X, Y, Z]): DataTensor[X, Y, Z] = {
    val colGroupedSlices = dm.dim2Indices.map(idx => dm.getCaCellsInDim2(idx))

    val idUpdates = regroupIds(colGroupedSlices)

    if (idUpdates.isEmpty) {
      //Nothing changed we can return the old one
      dm
    } else {
      val updates = idUpdates.get
      val newRowData = new Array[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]](dm.nrOf1DGroups)
      for (rowGroupIdx <- dm.dim1Indices) {
        var currentIdxShift = 0
        val newColData = new Array[IndexedSeq[CaCell[X, Y, Z]]](updates.size)
        for (colGroupIdx <- dm.dim2Indices) {
          updates.get(colGroupIdx).fold {
            currentIdxShift += 1 //This segment is empty after regrouping
          } { x =>
            newColData(colGroupIdx - currentIdxShift) = x.fold {
              dm.caCells(rowGroupIdx)(colGroupIdx) //No updates
            } { newIds =>
              dm.caCells(rowGroupIdx)(colGroupIdx).map { gt =>
                gt.copyWithNewDim2(dim2Ids = newIds)
              }
            }
          }
        }
        newRowData(rowGroupIdx) = newColData
      }
      dm.copyWithNewCaCells(newRowData)
    }
  }
}
