package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor}

/**
  * Implements the regroup operation for tubes
  */
object ReGroupTubes extends  AbstractRegroup{

  protected def determineEntryMatrixSize[X, Y, Z](ca: CaCell[X, Y, Z]): Int = ca.matrixSizeForDim3Idx

   def performOperation[X, Y, Z](dm: DataTensor[X, Y, Z]): DataTensor[X, Y, Z] = {
    val tubeGroupedSlices = dm.dim3Indices.map(idx => dm.getCaCellsInDim3(idx))
    val idUpdates = regroupIds(tubeGroupedSlices)

    if (idUpdates.isEmpty) {
      //Nothing changed we can return the old one
      dm
    } else {
      val updates = idUpdates.get
      val newRowData = new Array[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]](dm.nrOf1DGroups)
      for {rowGroupIdx <- dm.dim1Indices} {
        val newColData = new Array[IndexedSeq[CaCell[X, Y, Z]]](dm.nrOf2DGroups)
        for {colGroupIdx <- dm.dim2Indices} {
          var currentIdxShift = 0
          val newTubeData = new Array[CaCell[X, Y, Z]](updates.size)
          for {tubeGroupIdx <- dm.dim3Indices} {
            updates.get(tubeGroupIdx).fold {
              currentIdxShift += 1 //This segment is empty after regrouping
            } { x =>
              newTubeData(tubeGroupIdx - currentIdxShift) = x.fold {
                dm.caCells(rowGroupIdx)(colGroupIdx)(tubeGroupIdx) //No updates
              } { newIds =>
                dm.caCells(rowGroupIdx)(colGroupIdx)(tubeGroupIdx).copyWithNewDim3(dim3Ids = newIds)
              }
            }
          }
          newColData(colGroupIdx) = newTubeData
        }
        newRowData(rowGroupIdx) = newColData
      }
      dm.copyWithNewCaCells(newRowData)
    }
  }
}
