package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor}


/**
  * Split operation for columns
  */
object SplitCol extends  AbstractSplit {

  protected def performGroupSplit[X, Y, Z](dm: DataTensor[X, Y, Z], splitSeq: Seq[Int]): DataTensor[X, Y, Z] = {
    if (splitSeq.isEmpty) {
      dm
    } else {
      val splitGroup = splitSeq.head
      val groupToBeSplit = dm.getCaCellsInDim2(splitGroup)

      val (s1, s2) = determineNewGroupIds(groupToBeSplit)

      if (s2.nonEmpty && s1.nonEmpty) {
        val newRowData = new Array[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]](dm.nrOf1DGroups)
        for {
          rowIdx <- dm.dim1Indices
        } {
          val newLastIndex = dm.nrOf2DGroups
          val newColData = new Array[IndexedSeq[CaCell[X, Y, Z]]](newLastIndex + 1)
          for {
            colIdx <- dm.dim2Indices
          } {
            if (colIdx != splitGroup) {
              newColData(colIdx) = dm.caCells(rowIdx)(colIdx)
            } else {
              val old = dm.caCells(rowIdx)(colIdx)
              newColData(colIdx) = old.map { gt =>
                gt.copyWithNewDim2(dim2Ids = s1)
              }
              newColData(newLastIndex) = old.map { gt =>
                gt.copyWithNewDim2(dim2Ids = s2)
              }
            }
          }
          newRowData(rowIdx) = newColData
        }

        dm.copyWithNewCaCells(newRowData)

      } else {
        //We test the next one
        performGroupSplit(dm, splitSeq.tail)
      }
    }
  }

  protected def determineEntryMatrixSize[X, Y, Z](ca: CaCell[X, Y, Z]): Int = ca.matrixSizeForDim2Idx

  def indicesOfMaxEncoding[X, Y, Z](dm: DataTensor[X, Y, Z]): Seq[Int] = {
    (0 until dm.nrOf2DGroups).view.map {
      idx =>
        val slice = dm.getCaCellsInDim2(idx)
        slice.entropyPerSlice -> idx
    }.sortBy(_._1 * -1).map(_._2) //.take(1)
  }
}
