package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor}


/**
  * Split operation for rows
  */
object SplitRow extends  AbstractSplit {


  protected def performGroupSplit[X, Y, Z](dm: DataTensor[X, Y, Z], splitSeq: Seq[Int]): DataTensor[X, Y, Z] = {
    if (splitSeq.isEmpty) {
      dm
    } else {
      val splitGroup = splitSeq.head
      val groupToBeSplit = dm.getCaCellsInDim1(splitGroup)

      val (s1, s2) = determineNewGroupIds(groupToBeSplit)

      if (s2.nonEmpty && s1.nonEmpty) {
        val newLastIndex = dm.nrOf1DGroups
        val newData = new Array[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]](newLastIndex + 1)
        for {
          rowIdx <- dm.dim1Indices
        } {
          if (rowIdx != splitGroup) {
            newData(rowIdx) = dm.caCells(rowIdx)
          } else {
            newData(rowIdx) = dm.caCells(rowIdx).map(_.map { gt =>
              gt.copyWithNewDim1(dim1Ids = s1)
            })
            newData(newLastIndex) = dm.caCells(rowIdx).map(_.map { gt =>
              gt.copyWithNewDim1(dim1Ids = s2)
            })
          }
        }
        dm.copyWithNewCaCells(newData)
      } else {
        //We test the next one
        performGroupSplit(dm, splitSeq.tail)
      }
    }
  }

  protected def determineEntryMatrixSize[X, Y, Z](ca: CaCell[X, Y, Z]): Int = ca.matrixSizeForDim1Idx

  def indicesOfMaxEncoding[X, Y, Z](dm: DataTensor[X, Y, Z]): Seq[Int] = {
    (0 until dm.nrOf1DGroups).view.map {
      idx =>
        val slice = dm.getCaCellsInDim1(idx)
        slice.entropyPerSlice -> idx
    }.sortBy(_._1 * -1).map(_._2) // .take(1)
  }
}
