package subcluens.operations

import subcluens.datastructure.{CaCell, DataTensor}

/**
  * Split operation for tubes
  */
object SplitTube extends  AbstractSplit {

  def performGroupSplit[X, Y, Z](dm: DataTensor[X, Y, Z], splitSeq: Seq[Int]): DataTensor[X, Y, Z] = {
    if (splitSeq.isEmpty) {
      dm
    } else {
      val splitGroup = splitSeq.head
      val groupToBeSplit = dm.getCaCellsInDim3(splitGroup)


      val (s1, s2) = determineNewGroupIds(groupToBeSplit)

      if (s2.nonEmpty && s1.nonEmpty) {
        val newRowData = new Array[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]](dm.nrOf1DGroups)
        for {
          rowIdx <- dm.dim1Indices
        } {
          val newColData = new Array[IndexedSeq[CaCell[X, Y, Z]]](dm.nrOf2DGroups)
          for {
            colIdx <- dm.dim2Indices
          } {
            val newLastIndex = dm.nrOf3DGroups
            val newTubeData = new Array[CaCell[X, Y, Z]](newLastIndex + 1)
            for {
              tubeIdx <- dm.dim3Indices
            } {
              if (tubeIdx != splitGroup) {
                newTubeData(tubeIdx) = dm.getCaCells(rowIdx, colIdx, tubeIdx)
              } else {
                val old = dm.getCaCells(rowIdx, colIdx, tubeIdx)
                newTubeData(tubeIdx) = {
                  old.copyWithNewDim3(dim3Ids = s1)
                }
                newTubeData(dm.nrOf3DGroups) = {
                  old.copyWithNewDim3(dim3Ids = s2)
                }
              }
            }
            newColData(colIdx) = newTubeData
          }
          newRowData(rowIdx) = newColData
        }
        dm.copyWithNewCaCells(newRowData)
      } else {
        //We test the next one
        performGroupSplit(dm, splitSeq.tail)
      }
    }
  }

  protected def determineEntryMatrixSize[X, Y, Z](ca: CaCell[X, Y, Z]): Int = ca.matrixSizeForDim3Idx

  def indicesOfMaxEncoding[X, Y, Z](dm: DataTensor[X, Y, Z]): Seq[Int] = {
    (0 until dm.nrOf3DGroups).view.map {
      idx =>
        val slice = dm.getCaCellsInDim3(idx)
        slice.entropyPerSlice -> idx
    }.sortBy(_._1 * -1).map(_._2) //.take(1)
  }
}
